import math
import xlrd
from xlrd.biffh import XLRDError
import xlsxwriter
import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import curve_fit
from scipy import stats
from matplotlib.figure import Figure
import copy
import warnings
from scipy.optimize import OptimizeWarning
import io
import matplotlib as mpl


'''
core program
'''

def double_exp_func(x, A1, k1, A2, k2, y0):
    #print(A1, k1, A2, k2, y0, flush = True)
    return -A1 * np.exp(-k1 * x) - A2 * np.exp(-k2 * x) + y0

def single_exp_func(x, A1, k1, y0):
    #print(A1, k1, y0, flush = True)
    return -A1 * np.exp(-k1 * x) + y0

def lin(x, a, b):
    return a * x + b

def keys_to_index(key):
    Ch = key[0]
    j = ord(Ch)-ord('A') + 1 
    if len(key)==2:
        N = key[1]
    if len(key)==3:
        N = key[1:3]
    return j, int(N)
    
def index_to_keys(i,  j):
    y = str(chr(ord('A')-1+j))
    key = '{}{}'.format(y, i)
    return key

class FitterProgram:
    def __init__(self):
        self.imported_data_CFX = []
        self.imported_data_AriaMx = {}
        self.raw = {}
        self.lag_splitted = {}
        self.double_fitted = {}
        self.single_fitted = {}
        self.result = {}
        self.calib_all = {} #contains all the used calibration points
        self.calib_eff = {} #contains the well-behaved calibration points only (useful ones)
        self.outer = {} #contains the outliers
        self.zero = {} #contains which did not converge when fit (close to zero)
        #self.scanning_frequency = None
        #self.x_sec = []
        self.x_minutes = []
        self.average_k2 = None
        self.calAmplitude = []
        self.x_pmol = []
        self.slope = 0
        self.intercept = 0
        self.R2 = None
        self.is_calibration_ready = False
        self.split_j = 0
        self.runtwarn = []
        self.optwarn = []
        self.min_cal = None
        self.max_cal = None
        self.outer_results = []
        self.empty_wells = []
        self.Akaike_single = []
        self.AIC_single = None #This value continuously vary upon fitting the different keys
        self.equal_k1_k2 = []
        self.wrong_k1 = []
        self.wrong_k2 = []
        self.negative_A1 = []
        self.close_to_zero = []
        self.A1_high_error = []
        self.param_high_error = []

        self.fig = Figure()
        self.ax = self.fig.add_subplot(1,2,1)
        self.ax2 = self.fig.add_subplot(1,2,2)
        self.fig.subplots_adjust(wspace = 0.3)
      
    def clear_subplot(self, subplot_index):
        if subplot_index == 1:
            self.ax.clear()
        elif subplot_index == 2:
            self.ax2.clear()
    def clear_everything(self):
        self.imported_data_CFX.clear()
        self.imported_data_AriaMx.clear()
        self.raw.clear()
        self.lag_splitted.clear()
        self.double_fitted.clear()
        self.single_fitted.clear()
        self.result.clear()
        self.calib_all.clear()
        self.calib_eff.clear() 
        self.outer.clear()
        self.zero.clear()
        #self.scanning_frequency = None
        #self.x_sec.clear()
        self.x_minutes.clear()
        self.average_k2 = None
        self.calAmplitude.clear()
        self.x_pmol.clear()
        self.slope = None
        self.intercept = 0
        self.R2 = None
        self.is_calibration_ready = False
        self.split_j = 0
        self.runtwarn = []
        self.optwarn = []
        self.min_cal = None
        self.max_cal = None
        self.outer_results = []
        self.empty_wells = []
        self.Akaike_single = []
        self.AIC_single = None
        self.equal_k1_k2 = []
        self.wrong_k1 = []
        self.wrong_k2 = []
        self.negative_A1 = []
        self.close_to_zero = []
        self.A1_high_error = []
        self.param_high_error = []

        self.clear_subplot(1)
        self.clear_subplot(2)
        
    def filereader_CFX(self, file_path):
        my_imported_excel = xlrd.open_workbook(file_path)
        sheet = my_imported_excel.sheet_by_index(0)
    
        self.imported_data_CFX = [[sheet.cell_value(r, c) for r in range(sheet.nrows)] for c in range(1, sheet.ncols)]
        
        file_OK = False
        for elements in self.imported_data_CFX:
            if 'Cycle' in elements:
                file_OK = True
        if not file_OK:
            raise Exception('The sheet does not contain the desired headings (e.g. Cycle in the B1 field)')
    
    def process_data_CFX(self, scan_freq):
        
        for elements in self.imported_data_CFX:
            if elements[0] == 'Cycle' and 'Time' not in self.raw.keys():
                time = elements[1:len(elements)]
                self.raw['Time'] = [i*int(scan_freq) for i in time]
            elif elements[0] == 'Cycle' and 'Time' in self.raw.keys():
                time = elements[1:len(elements)]
                self.raw['Time'].extend([i*int(scan_freq) + self.raw['Time'][-1] for i in time])
            elif elements[0].strip() != 'Cycle' and elements[0].strip() not in self.raw.keys():
                self.raw[elements[0].strip()] = elements[1:len(elements)]
            elif elements[0].strip() != 'Cycle' and elements[0].strip() in self.raw.keys():
                self.raw[elements[0].strip()].extend(elements[1:len(elements)])

        
    def filereader_AriaMx(self, file_path):
        my_imported_excel = xlrd.open_workbook(file_path)
        number_of_sheets = my_imported_excel.nsheets

        appropriate_sheet = None

        for i in range(number_of_sheets):
            sheet = my_imported_excel.sheet_by_index(i)
            if sheet.cell_value(0, 0) == 'Amplification Plots':
                appropriate_sheet = my_imported_excel.sheet_by_index(i)
        if appropriate_sheet == None:
            raise Exception('The file does not contain the desired sheet (\'Amplification Plots\' in the A1 field)')

        cycles = appropriate_sheet.cell_value(appropriate_sheet.nrows-1, 1) #row, column
        
        self.imported_data_AriaMx['Cycle'] = [appropriate_sheet.cell_value(r, 1) for r in range(3, int(cycles) + 3)]
        
        number_of_wells = (appropriate_sheet.nrows-1)/(cycles + 2)        

        for jumper in range(2, ((int(cycles)+2)*int(number_of_wells)), (int(cycles)+2)):    
            key_cell_content = appropriate_sheet.cell_value(jumper, 0).split(",")
            key = key_cell_content[0]
            self.imported_data_AriaMx[key] = [appropriate_sheet.cell_value(jumper+1, 2) for jumper in range(jumper, jumper+int(cycles))]

        #print(self.imported_data_AriaMx)
        
##        file_OK = False
##        for elements in self.imported_data_CFX:
##            if 'Cycle' in elements:
##                file_OK = True
##        if not file_OK:
##            raise Exception('The sheet does not contain the desired headings (e.g. Cycle in the B1 field)')
##    
    def process_data_AriaMx(self, scan_freq):
        
        for elements in self.imported_data_AriaMx:
            if elements == 'Cycle' and 'Time' not in self.raw.keys():
                time = self.imported_data_AriaMx[elements]
                self.raw['Time'] = [i*int(scan_freq) for i in time]
            elif elements == 'Cycle' and 'Time' in self.raw.keys():
                time = self.imported_data_AriaMx[elements]
                self.raw['Time'].extend([i*int(scan_freq) + self.raw['Time'][-1] for i in time])
            elif elements != 'Cycle' and elements not in self.raw.keys():
                self.raw[elements] = self.imported_data_AriaMx[elements]
            elif elements != 'Cycle' and elements in self.raw.keys():
                self.raw[elements].extend(self.imported_data_AriaMx[elements])
        
        
    def filereader_QuantStudio(self, file_path):
        my_imported_excel = xlrd.open_workbook(file_path)
        try:
            sheet = my_imported_excel.sheet_by_name('Multicomponent Data')
        except XLRDError:
            raise Exception('Missing data: this file does not contain the desired Multicomponent Data sheet\nPlease reexport your QuantStudio measurement file ticking Multicomponent Data also!')

        FAM_position_col = None
        FAM_position_row = None
        for row in range(sheet.nrows):
            for col in range(sheet.ncols):
                if sheet.cell_value(row, col) == 'FAM':
                    FAM_position_col = col
                    FAM_position_row = row

        Cycle_position_col = None
        Cycle_position_row = None
        for row in range(sheet.nrows):
            for col in range(sheet.ncols):
                if sheet.cell_value(row, col) == 'Cycle':
                    Cycle_position_col = col
                    Cycle_position_row = row


        if FAM_position_col == None:
            raise Exception('FAM position not found!\nPlease ensure that the data originates from FAM channel and FAM label is above the results column')
       
        start_cycle = FAM_position_row + 1
        stop_cycle = start_cycle + 96
        for rows in range(start_cycle, stop_cycle):
            key = sheet.cell_value(rows, Cycle_position_col - 1).strip()
            self.raw[key] = []
        self.raw['Cycle'] = []
        for cycles in range(0, int((sheet.nrows - FAM_position_row ) / 96)):
            self.raw['Cycle'].append(sheet.cell_value(start_cycle, Cycle_position_col))
            for rows in range(start_cycle, stop_cycle):
                key = sheet.cell_value(rows, Cycle_position_col - 1).strip()
                if sheet.cell_value(rows, FAM_position_col) == '':
                    self.raw.pop(key, None)
                else:
                    try:
                        self.raw[key].append(sheet.cell_value(rows, FAM_position_col))
                    except Exception as exc:
                        raise Exception('Data available to non-existing well: '+str(exc))
                   
            start_cycle = start_cycle + 96
            stop_cycle = stop_cycle + 96
            
    def process_data_QS(self, time_list):
        try:
            if len(time_list) == 1:
                if time_list[0][0].value == '':
                    raise Exception('You must type the time intervals of plate reading!')
                else:
                    self.raw['Time'] = [elements*int(time_list[0][0].value) for elements in self.raw['Cycle']]
            else:
                
                for i in range(0, len(time_list)):
                    if time_list[i][0].value == '' or time_list[i][1].value == '':
                        raise Exception('You must fill all of the input fields!')
                        
                self.raw['Time'] = [elements*int(time_list[0][0].value) for elements in self.raw['Cycle'][0:int(time_list[0][1].value)]]

                sum_cycle = 0
                for i in range(1, len(time_list)):
                    sum_cycle = sum_cycle + int(time_list[i-1][1].value)
                    self.raw['Time'].extend([(self.raw['Time'][-1] + (elements-sum_cycle)*int(time_list[i][0].value)) for elements in self.raw['Cycle'][sum_cycle:(sum_cycle + int(time_list[i][1].value))]])

                    if i == len(time_list)-1:
                        if (sum_cycle + int(time_list[i][1].value)) != len(self.raw['Cycle']):
                            raise Exception('The number of the cycles does not match with data length')
        except ValueError:
            raise Exception('You must type numbers (integers) into the input fields')

               
    def draw_curve(self, ax, keys):
        if len(self.double_fitted) != 0:
            x_mins = []
            x_mins.clear()
            
            data_length = int(len(self.lag_splitted[keys]))
            time = copy.deepcopy(self.raw['Time'])
            time_length = int(len(time))
            starting_index = time_length - data_length
            x = np.array(time[starting_index:])
            
            #x = np.array(self.raw['Time'])

            for elements in x:
                x_mins.append(elements/60) #convert to minutes
        
            #x_mins = np.array(self.x_minutes)
            #print(x_mins)

            #print(len(x_mins))
            
            
            ax.set(xlabel='Time (min)', ylabel='Fluorescence',title='Double exponential fit results')
            try:
                if keys != "Cycle":
                
                    y = self.lag_splitted[keys]
                    #print(len(y))
                
                curvey = double_exp_func(x,self.double_fitted[keys][0],self.double_fitted[keys][1],self.double_fitted[keys][2],self.double_fitted[keys][3],self.double_fitted[keys][4])
                ax.plot(x_mins, curvey, 'lemonchiffon')
                ax.scatter(x_mins,y, c='none', edgecolors='salmon', s = 22, label='{}'.format(keys))
            except KeyError:
                    pass

    def draw_calib(self, ax):
        if len(self.x_pmol) != 0:
            
            x = np.array(self.x_pmol)
            
            curvey2 = lin(x,self.slope,self.intercept) 
            self.clear_subplot(2)

            if self.intercept >= 0:
                ax.plot(self.x_pmol, curvey2, 'limegreen', label='{:.2f}x + {:.2f}, $R^2$= {:.3f}'.format(self.slope , self.intercept, self.R2))
            elif self.intercept < 0:
                ax.plot(self.x_pmol, curvey2, 'limegreen',
                         label='{:.2f}x – {:.2f}, $R^2$= {:.3f}'.format(self.slope, abs(self.intercept), self.R2))
            ax.scatter(self.x_pmol, self.calAmplitude, c='indianred', s = 22, label='data points')
            ax.legend(loc='best')
            ax.set(xlabel='dNTP (pmol)', ylabel='Amplitude',title='Calibration')

    def calib_magnifier(self, key):
        y = self.double_fitted[key][0]
        x = self.calib_eff[key]
        magnified = self.ax2.scatter(x, y, c='indianred', edgecolors='black', s = 100, label='data points')

    def curve_analysis(self):
               
        for key in self.raw:
            split_index = 0
            if key != "Time" and key !="Cycle":
                y = self.raw[key]
                for index in range(0, len(y)-5):

                    subtract_1 = y[index + 1] - y[index]
                    subtract_2 = y[index + 2] - y[index + 1]
                
                    if subtract_2 < subtract_1:
                        split_index = index
                        break
                
                self.lag_splitted[key] = copy.deepcopy(self.raw[key])

                for i in range(0, split_index):
            
                    self.lag_splitted[key].pop(0) #ez idáig jó!
                #print(key, self.lag_splitted[key]) #ez idáig jó, self.lag_splitted csak azt tartalmazza, amibõl levágtunk és jó adatmennyiséget

    def single_fitting(self, keys):

        #x = np.array(self.raw['Time'])

        #if keys != "Time" and keys !="Cycle":
        y = self.lag_splitted[keys]

        A_init = self.lag_splitted[keys][len(self.lag_splitted[keys])-1] #This is the last data point of the key, fitting iteration begins with it
        
        A_min = A_init #Try to find k_init value for A_init*2/3

        t_min = 1000
        for elements in self.raw[keys]:
            A_actual_min = abs(A_init * 2 / 3 * 2 / 3 - elements)
            if A_actual_min < A_min:
                A_min = A_actual_min
                k_min = self.raw[keys].index(elements) #k_min az indexe lesz az ominózus elemnek asszem, itt cserélendõ lesz
                t_min = self.raw['Time'][k_min]
                
        k_init = 1 / t_min
        
        iteration = 1200000000

        data_length = int(len(self.lag_splitted[keys]))
        time = copy.deepcopy(self.raw['Time'])
        time_length = int(len(time))
        starting_index = time_length - data_length
        x = np.array(time[starting_index:])
    
        try:
            popt, pcov = curve_fit(single_exp_func,x,y,p0=(A_init *2/3,k_init,A_init),maxfev=iteration)
        except RuntimeError:
            iteration = iteration * 1000
            popt, pcov = curve_fit(single_exp_func, x, y, p0=(A_init *2/3,k_init,A_init), maxfev=iteration)
        except RuntimeWarning:
           pass
        except:
            pass


        p1 = popt[0] # This is your A1
        p2 = popt[1] # This is your k1
        p3 = popt[2] # This is your y0

        residuals = y - single_exp_func(x,p1,p2,p3)
        rss_single = sum(residuals**2)
        fres = sum( (residuals**2)/single_exp_func(x,p1,p2,p3) ) # The chi-sqaure of your fit
        
        self.single_fitted[keys] = [p1, p2, p3, fres, pcov.tolist()]
        #print(keys, self.single_fitted[keys][0:3])
        
        k = 3 # of variables
        N = len(self.lag_splitted[keys])

        if N >= 120:
            self.AIC_single = 2*k + N*np.log(rss_single)
            #print('AIC_single', keys, self.AIC_single)
            #print('single_rss', rss_single)
        elif N < 120:
            self.AIC_single = 2*k + N*np.log(rss_single) + (2*k*(k+1)/(N-k-1))
            
    def double_fitting(self):

        if 'Time' not in self.raw.keys():
            raise Exception('No datafile has been read!')
        
        if self.is_calibration_ready == False:      #This is important when open a .dntp file, to avoid x duplication on the figure(13 sec * 13 sec)
            for elements in self.raw['Time']:
                self.x_minutes.append(elements/60) #convert to minutes
        
        #print(x)

        sum_k2 = 0
        for keys in self.lag_splitted:
            #if keys != "Time" and keys !="Cycle":
            y = self.lag_splitted[keys]
            A_init = self.lag_splitted[keys][len(self.lag_splitted[keys])-1]
            A_min = A_init
            t_min = 1000
            for elements in self.raw[keys]:
                A_actual_min = abs(A_init * 2 / 3 * 2 / 3 - elements)
                if A_actual_min < A_min:
                    A_min = A_actual_min
                    k_min = self.raw[keys].index(elements)
                    t_min = self.raw['Time'][k_min]
                    

            k_init = 1 / t_min
                
            iteration = 1200000000

            data_length = int(len(self.lag_splitted[keys]))
            time = copy.deepcopy(self.raw['Time'])
            time_length = int(len(time))
            starting_index = time_length - data_length
            x = np.array(time[starting_index:])
            
            try:
                with warnings.catch_warnings(record=True) as w:
                    popt, pcov = curve_fit(double_exp_func,x,y,p0=(A_init *2/3,k_init,A_init*1/3,k_init / 10,A_init),maxfev=iteration)
            except RuntimeError:
                iteration = iteration * 1000
                with warnings.catch_warnings(record=True) as w:
                    popt, pcov = curve_fit(double_exp_func, x, y, p0=(A_init *2/3,k_init,A_init*1/3,k_init / 10,A_init), maxfev=iteration)

            for i in w:
                if issubclass(i.category, OptimizeWarning):
                    self.optwarn.append(keys)
            for i in w:
                if issubclass(i.category, RuntimeWarning):
                    self.runtwarn.append(keys)
                 
##                   
##                except OptimizeWarning:
##                    self.optwarn.append(keys)

            p1 = popt[0] # This is your A1
            p2 = popt[1] # This is your k1
            p3 = popt[2] # This is your A2
            p4 = popt[3] # This is your k2
            p5 = popt[4] # This is your y0

            residuals = y - double_exp_func(x,p1,p2,p3,p4,p5)
            rss_double = sum(residuals**2)
            fres = sum( (residuals**2)/double_exp_func(x,p1,p2,p3,p4,p5) ) # The chi-sqaure of your fit
               
            if p2 > p4:
                self.double_fitted[keys] = [p1, p2, p3, p4, p5, fres, pcov.tolist()]
            elif p2 < p4:
                self.double_fitted[keys] = [p3, p4, p1, p2, p5, fres, pcov.tolist()]

            sum_k2 = sum_k2 + self.double_fitted[keys][3]

            self.draw_curve(self.ax, keys)


            k = 5 # of variables
            N = len(self.lag_splitted[keys])
            
            if N >= 200:
                AIC_double = 2*k + N*np.log(rss_double)
            elif N < 200 and N != 6:
                AIC_double = 2*k + N*np.log(rss_double) + (2*k*(k+1)/(N-k-1))
            elif N == 6:
                AIC_double = 2*k + N*np.log(rss_double)
            self.single_fitting(keys)

            if self.AIC_single < AIC_double:
                self.Akaike_single.append(keys)
            
            
    
        self.average_k2 = sum_k2 / len(self.double_fitted)
        #print('Akaike single', self.Akaike_single)          

    def calib_calc(self):
        self.calib_eff = copy.deepcopy(self.calib_all)
        self.calAmplitude.clear()
        self.x_pmol.clear()
        self.outer_results.clear()
        self.close_to_zero.clear()
        
        for keys in self.calib_all:
            try:
                if keys in self.Akaike_single:
                    self.calib_eff.pop(keys, None)
                    self.zero[keys] = self.calib_all[keys]
                elif self.double_fitted[keys][1] / self.double_fitted[keys][3] < 1.1:
                    self.calib_eff.pop(keys, None)
                    self.zero[keys] = self.calib_all[keys]
                    self.equal_k1_k2.append(keys)
                elif self.double_fitted[keys][1] / self.average_k2 < 2:
                    self.calib_eff.pop(keys, None)
                    self.zero[keys] = self.calib_all[keys]
                    self.wrong_k1.append(keys)
                elif self.double_fitted[keys][0] < 0:
                    self.calib_eff.pop(keys, None)
                    self.zero[keys] = self.calib_all[keys]
                    self.negative_A1.append(keys)
                elif keys in self.outer:
                    self.calib_eff.pop(keys, None)
                elif (pow(abs(self.double_fitted[keys][6][0][0]), (0.5))) > (self.double_fitted[keys][0]):
                    self.calib_eff.pop(keys, None)
                    self.zero[keys] = self.calib_all[keys]
                    self.A1_high_error.append(keys)
                    #print("A1 error:", self.A1_high_error)
                #elif self.double_fitted[keys][3] < pow(10, (-5)):
                    #self.calib_eff.pop(keys, None)
                    #self.zero[keys] = self.calib_all[keys]
                    #self.wrong_k2.append(keys)
                else:
                    self.calAmplitude.append(self.double_fitted[keys][0])
                    #print("calAmp:", self.calAmplitude)
            except KeyError as exc:
                self.calib_all.pop(keys, None)
                self.calib_eff.pop(keys, None)
                raise Exception('No data available to the ' + str(exc) + ' well. Please look at your imported xlsx file, if this well has any missing data.' )
            #print(keys, "zero", self.zero, "k12:", self.equal_k1_k2, "wring_k1", self.wrong_k1, "negA", self.negative_A1, "err", self.A1_high_error, "k2", self.wrong_k2)
        for elements in self.calib_eff:
            self.x_pmol.append(self.calib_eff[elements])
            #print("x_pmol:", self.x_pmol)
            
    

        try:
            if len(self.x_pmol)==0:
                raise Exception('The calibration wells were not specified or they were ejected by the built-in checkpoints of the software. If you specified the calibration wells with the values, please look at the Parameter sheet of the Excel file for the reason.')
            self.min_cal = min(self.x_pmol)
            self.max_cal = max(self.x_pmol)
            minimal_Amplitude_list = [keys for keys, value in self.calib_eff.items() if value == self.min_cal]
            total_Amp = self.raw[minimal_Amplitude_list[0]][-1] - self.raw[minimal_Amplitude_list[0]][0] #with this I select the total Amplitude for the least calibration point in order to filter out empty wells or wells with very low Amplitude
        
            y_values = self.calAmplitude
            slope, intercept, r_value, p_value, std_err = stats.linregress(self.x_pmol, y_values)
            self.slope = slope
            self.intercept = intercept
            self.R2 = r_value**2

            self.draw_calib(self.ax2)
        
        except:
            raise Exception('The calibration wells were not specified or they were ejected by the built-in checkpoints of the software. If you specified the calibration wells with the values, please look at the Parameter sheet of the Excel file for the reason.')
    
        

        """
        The following part calculates the data points 
        """
        
        for keys in self.double_fitted:
            if keys in self.Akaike_single:
                self.result[keys] = 'x'
            if self.double_fitted[keys][1] / self.double_fitted[keys][3] < 1.1:
                self.result[keys] = 'x'
                self.equal_k1_k2.append(keys)
            if self.double_fitted[keys][1] / self.average_k2 < 2:
                self.result[keys] = 'x'
                self.wrong_k1.append(keys)
            if self.double_fitted[keys][0] < 0:
                self.result[keys] = 'x'
                self.negative_A1.append(keys)
            if (pow(abs(self.double_fitted[keys][6][0][0]), (0.5))) > (self.double_fitted[keys][0]):
                self.result[keys] = 'x'
                self.A1_high_error.append(keys)
            if self.raw[keys][-1] - self.raw[keys][0] < total_Amp/2:
                self.result[keys] = 'x'
                self.empty_wells.append(keys)
            #if self.double_fitted[keys][3] < pow(10, (-5)):
             #   self.result[keys] = 'x'
            #    self.wrong_k2.append(keys)
            if keys not in self.calib_all and keys not in self.Akaike_single and keys not in self.equal_k1_k2 and keys not in self.wrong_k1 and keys not in self.negative_A1 and keys not in self.A1_high_error and keys not in self.empty_wells and keys not in self.wrong_k2:
                self.result[keys] = (self.double_fitted[keys][0] - intercept) / slope
                if self.result[keys] < 0:
                    self.result[keys] = 'x'
                    self.close_to_zero.append(keys)
            elif keys in self.calib_all: #defining here the calib all serves that for all of the calibration wells the calibration data is shown in the excel table, not an 'x'
                self.result[keys] = self.calib_all[keys] 
            if self.result[keys] != 'x':
                if float(self.result[keys]) < float(self.min_cal) or float(self.result[keys]) > float(self.max_cal):
                    #print(self.result[keys], self.min_cal, self.max_cal)
                    self.outer_results.append(keys)
            if (pow(abs(self.double_fitted[keys][6][1][1]), (0.5))) > self.double_fitted[keys][1] or (pow(abs(self.double_fitted[keys][6][2][2]), (0.5))) > self.double_fitted[keys][2] or (pow(abs(self.double_fitted[keys][6][3][3]), (0.5))) > self.double_fitted[keys][3] or (pow(abs(self.double_fitted[keys][6][4][4]), (0.5))) > self.double_fitted[keys][4]:
                self.param_high_error.append(keys) 
        self.is_calibration_ready = True
       
        
    """
    Export data to excel 
    """
    def export_to_Excel(self, file_name):
        fitted_data_workbook = xlsxwriter.Workbook(file_name, {'strings_to_numbers': True})
        Res_sheet = fitted_data_workbook.add_worksheet('Results')
        superscript = fitted_data_workbook.add_format({'font_script': 1})
        subscript = fitted_data_workbook.add_format({'font_script': 2})

##        print('outer', self.outer)
##        print('zero', self.zero)
##        print('outer', self.outer_results)
##        print('empty', self.empty_wells)
##        print('Akaike', self.Akaike_single)
##        print('equal', self.equal_k1_k2)
##        print('wrong k1', self.wrong_k1)
##        print('negative A', self.negative_A1)
##        print('below_zero', self.close_to_zero)
##        print('A1 error', self.A1_high_error)
##        print('param error', self.param_high_error)
        
        
        row = 0
        col = 0
        for keys in self.result:
            
            Res_sheet.write(row, col, '{}'.format(keys))
            Res_sheet.write(row, col + 1, '{}'.format(self.result[keys]))
            Res_sheet.write(row, col + 2, 'pmol')
            if keys in self.outer_results:
                Res_sheet.write(row, col + 3, 'out of calibration range')
            #if keys in self.outer_results and keys in self.calib_all:
                #Res_sheet.write(row, col + 3, 'removed from calibration')
            if keys in self.param_high_error and self.result[keys] != 'x' and keys not in self.calib_all:
                Res_sheet.write(row, col + 3, 'high fitting error of some parameters')
            
            row += 1

        export_fig_cal, export_ax_cal = plt.subplots(1, 1)


        export_ax_cal.ticklabel_format(style='sci', axis='y', scilimits=(0,0))
        export_ax_cal.yaxis.major.formatter._useMathText = True
        

        self.draw_calib(export_ax_cal)

        imgdata=io.BytesIO()
        export_fig_cal.savefig(imgdata, format='png')
        Res_sheet.insert_image(1, 5,'',{'image_data': imgdata})
        
        Param_sheet = fitted_data_workbook.add_worksheet('All_parameters')
        Param_sheet.freeze_panes(1, 0)
        row = 1
        col = 0
        Param_sheet.write_rich_string(0, 1, 'A', subscript, '1')
        Param_sheet.write_rich_string(0, 2, 'A', subscript, '1', ' se')
        Param_sheet.write_rich_string(0, 3, 'k', subscript, '1')
        Param_sheet.write_rich_string(0, 4, 'k', subscript, '1', ' se')
        
        Param_sheet.write_rich_string(0, 5, 'A', subscript, '2')
        Param_sheet.write_rich_string(0, 6, 'A', subscript, '2', ' se')
        Param_sheet.write_rich_string(0, 7, 'k', subscript, '2')
        Param_sheet.write_rich_string(0, 8, 'k', subscript, '2', ' se')
        Param_sheet.write_rich_string(0, 9, 'y', subscript, '0')
        Param_sheet.write_rich_string(0, 10, 'y', subscript, '0', ' se')
        Param_sheet.write(0, 11, 'Chi-square')
        Param_sheet.write(0, 13, 'Failure')
        
        export_fig_fits, export_ax_fits = plt.subplots(1, 1)
        
        export_ax_fits.ticklabel_format(style='sci', axis='y', scilimits=(0,0))
        export_ax_fits.yaxis.major.formatter._useMathText = True
        
        
        for keys in self.double_fitted:
            self.draw_curve(export_ax_fits, keys)
            
        imgdata=io.BytesIO()
        export_fig_fits.savefig(imgdata, format='png')
        Param_sheet.insert_image(3, 18,'',{'image_data': imgdata})

        for keys in self.double_fitted:

            Param_sheet.write(row, col, '{}'.format(keys))
            Param_sheet.write(row, col + 1, '{}'.format(self.double_fitted[keys][0]))
            Param_sheet.write(row, col + 3, '{}'.format(self.double_fitted[keys][1]))
            Param_sheet.write(row, col + 5, '{}'.format(self.double_fitted[keys][2]))
            Param_sheet.write(row, col + 7, '{}'.format(self.double_fitted[keys][3]))
            Param_sheet.write(row, col + 9, '{}'.format(self.double_fitted[keys][4]))
            Param_sheet.write(row, col + 11, '{}'.format(self.double_fitted[keys][5]))
            Param_sheet.write(row, col + 2, '{}'.format(pow(abs(self.double_fitted[keys][6][0][0]), (0.5))))
            Param_sheet.write(row, col + 4, '{}'.format(pow(abs(self.double_fitted[keys][6][1][1]), (0.5))))
            Param_sheet.write(row, col + 6, '{}'.format(pow(abs(self.double_fitted[keys][6][2][2]), (0.5))))
            Param_sheet.write(row, col + 8, '{}'.format(pow(abs(self.double_fitted[keys][6][3][3]), (0.5))))
            Param_sheet.write(row, col + 10, '{}'.format(pow(abs(self.double_fitted[keys][6][4][4]), (0.5))))
            
            if keys in self.runtwarn:
                Param_sheet.write(row, col + 12, 'the least-squares minimization failed')




            if keys in self.param_high_error:
                Param_sheet.write(row, col + 13, 'high fitting error of some parameters')
            if keys in self.empty_wells:
                Param_sheet.write(row, col + 13, 'very low total Amplitude')
            if keys in self.negative_A1:
                Param_sheet.write_rich_string(row, col + 13, 'negative A', subscript, '1')
            if keys in self.close_to_zero:
                Param_sheet.write(row, col + 13, 'out of calibration range')
            if keys in self.wrong_k2:
                Param_sheet.write_rich_string(row, col + 13, 'wrong k', subscript, '2')
            if keys in self.wrong_k1:
                Param_sheet.write_rich_string(row, col + 13, 'k', subscript, '1', ' is in the range of typical k', subscript, '2', '. If the empty wells of the plate contained Fluorescence data, please delete them and do the analysis again to get result of this well.')
            if keys in self.equal_k1_k2:
                Param_sheet.write_rich_string(row, col + 13, 'equal k', subscript, '1', ' and k', subscript, '2', ' observed')
            if keys in self.A1_high_error:
                Param_sheet.write_rich_string(row, col + 13, 'high fitting error of A', subscript, '1', ' parameter' )
            if keys in self.Akaike_single:
                Param_sheet.write(row, col + 13, 'no specific dNTP incorporation phase observed')
            row += 1
##        print("high_error", self.param_high_error)
##        print("empty", self.empty_wells)
##        print("negA", self.negative_A1)
##        print("zeor", self.close_to_zero)
##        print('k2', self.wrong_k2)
##        print('k1', self.wrong_k1)
##        print("k1k2", self.equal_k1_k2)
##        print("A1", self.A1_high_error)
##        print("single", self.Akaike_single)

        fitted_data_workbook.close()
